
//
//    (feb'07 19-02:07) - version v.07a
//                                                                                 the FLEUR project
// -------------------------------------------------------------------------------------------------
//                                                                                anton.empl@cern.ch
// --- FLUKA user back-end project -Fbe-
//     see header file (FbeUsrbinFile.h) for comments
//


#include "FbeUsrbinFile.h"

ClassImp(FbeUsrbinFile)



//
//
// -- here's our constructor
//    all the work is here -- VERY SPECIFIC TO USRBIN FILE STRUCTURE
//    open our binary input file (block ascii) and check what kind of file we are looking at
//    read the header stuff and see if this is a 'summed' usrbin data file in which case there's extra
//    statistics/error information per binning available - that extra information is found at the end
//    of the file, it was appended
//    binary infomation is interpreted regardless of endianess (tested linux/compaq/sun data files)
//    next extract 'per-file' information (data and statistics) keep file pointers at first for the
//    actual data so we can read-on-demand and have a more efficient byte swapping
//
//

FbeUsrbinFile::FbeUsrbinFile(const char *aName) : FbeIO(aName)
{
    int i,j,bytesRead,bytes,ret,position, NB=0,IB;
    char buf[256],*summed=NULL;

    _withStatistics= kFALSE;


//    try and learn what FLUKA input file we are dealing with, the checks are performed by the
//    identify() method - all this strongly relies on the actual specifics how FLUKA writes these
//    files - at this point we mainly work with USRBIN data files, only binary is implemented
//    but either endianess can be dealt with - _status will inform us about the degree of success
//    in opening the file
// ---------------------------------------------------------------------------------------------------

    identify();


    DEBUG(printf(" -debug-> FbeUsrbinFile _status %i\n",_status);)
    if (_status) return;
// ---------------------------- for now we assume here it's a USRBIN data file


//    if there's no date we know we will find statistics ie statistical error information per
//    binning at the end of the file - the file was written in FORTRAN so we need to read off record
//    length specifiers (twice, before and after the actual bytes) they are found in  bytes  and
//    tell us how many bytes to read
// ---------------------------------------------------------------------------------------------------

    _fd= open(_filename,O_RDONLY);


//    read throught the actual usrbin file header information, clean up the strings from the back
//    swapping of bytes is taken care of by the read() method(s)
// ---------------------------------------------------------------------------------------------------

    bytesRead= read(_fd,&bytes,4);

    bytesRead= read(_fd,buf,80);
    i=0; while (i<80 && buf[i]==' ') i++;
    j=79; while (j>i && buf[j]==' ') j--;
    strncpy(_title,&buf[i],j-i+1); _title[j-i+1]= '\0';
    bytesRead= read(_fd,buf,32);  i= bytesRead-1; while (buf[i]==' ') i--; strncpy(_date,buf,i+1);

    summed= strstr(_date," Sum file ");

    bytesRead= read(_fd,&_weightOfPrimary,4);
    bytesRead= read(_fd,&_nrOfEvents,4);
    bytesRead= read(_fd,&bytes,4);

// -----------------------------------------

//    loop over binnings, we don't know how many of those initially
//    read some variables a little 'optimized' ie need to cast some int onto float later
//    keep only file-pointers for the actual data/statistics information at first - read on demand
//    the file pointers are absolute pointers at this point
//    push all of this into another class FebUsrbin kept in a TObjArray
// ---------------------------------------------------------------------------------------------------

    char name[10];
    int nr,type,score,ctr=0;
    float arr[16];


    while (1) {

        bytesRead= read(_fd,&bytes,4);
        if (bytesRead<4) break;

        if (summed && bytes==STATISTICAL_ERROR_AVAILABLE) {
          bytesRead= read(_fd,buf,10); buf[10]= '\0';
          bytesRead= read(_fd,&IB,4);
          bytesRead= read(_fd,&bytes,4);
          _withStatistics= kTRUE;
          break;
          }

// --------------------------------------- header ---
        bytesRead= read(_fd,&nr,4);
        bytesRead= read(_fd, name,10); name[10]= '\0';
        bytesRead= read(_fd,&type,4);
        bytesRead= read(_fd,&score,4);

//      NB, TITUSB(NB), JTUHLP, IDUSBN(NB),
//      XLOW(NB), XHIGH(NB), NXBIN(NB), DXUSBN(NB) 
//      YLOW(NB), YHIGH(NB), NYBIN(NB), DYUSBN(NB) 
//      ZLOW(NB), ZHIGH(NB), NZBIN(NB), DZUSBN(NB) 
//      LNTZER(NB), BKUSBN(NB), B2USBN(NB), TCUSBN(NB) 
// ---------------------------------------------------
        bytesRead= read(_fd,arr,64);

        bytesRead= read(_fd,&bytes,4);

// --------------------------------------- data ------
        bytesRead= read(_fd,&bytes,4);
        position= lseek(_fd,0,SEEK_CUR);
        ret= lseek(_fd,bytes,SEEK_CUR);
        _binning.Add(new FbeUsrbin(ctr++,nr,name,type,score,arr,position,bytes));
        bytesRead= read(_fd,&bytes,4);

        NB++; }

    _nrOfBinnings= NB;
    //printf("%i\n",NB);

    if (_withStatistics) {

//    brief check on consitency, otherwise we read through the statistical error estimates at the
//    end of our summed FLUKA usrbin data file - and assume they are aligned
// ---------------------------------------------------------------------------------------------------

      if (NB!=IB) { printf(" -debug-> inconsistent statistics info (%i/%i)",NB,IB); _status= 4; }

      for (i=0;i<IB;i++) { bytesRead= read(_fd,&bytes,4);
                           position= lseek(_fd,0,SEEK_CUR);
                           ret= lseek(_fd,bytes,SEEK_CUR);
                           ((FbeUsrbin*)_binning[i])->addStatistics(position,bytes);
                           bytesRead= read(_fd,&bytes,4);
                         }
      }


} // FbeUsrbinFile



//
//
// -- here's our destructor
//    free out the binnings which are contained in the FbeUsrbin class stored as TObjArray
//

FbeUsrbinFile::~FbeUsrbinFile()
{

    if (_fd>0) close(_fd);

    DEBUG(printf(" -debug-> ~FbeUsrbinFile() -about to call .Delete on _binning-\n");)
    _binning.Delete();

} // ~FbeUsrbinFile



//
// -- with these 2 methods we return a binning to the outside world
//    and actually read the data from file
//


FbeUsrbin* FbeUsrbinFile::getABinning(int aBin)
{
if (aBin>_binning.GetLast()) return (0);
  else { FbeUsrbin *B= (FbeUsrbin*)_binning[aBin]; return (B); }
}



void FbeUsrbinFile::readData(FbeUsrbin *B)
{

//    ok, read in data for this binning and estimate on statistical errors if available
//    maybe not the cleanest design since the two Fbe classes are much intertwined
// ---------------------------------------------------------------------------------------------------

if (lseek(_fd,B->_from,SEEK_SET)<0) /*---need to catch error here---*/ {;}
  else {
         float *arr= (float*)malloc(B->_howManyBytes);
         int bytesRead= read(_fd,arr,B->_howManyBytes);
         B->_data.Adopt(bytesRead/4,arr);

         DEBUG(printf(" -debug-> read from file %3i\n",B->indexof());)

         if (B->_withStatisticalErrorEstimate) { if (lseek(_fd,B->_Sfrom,SEEK_SET)<0) /*---*/ {;}
                                                   else {
                                                          arr= (float*)malloc(B->_ShowManyBytes);
                                                          bytesRead= read(_fd,arr,B->_ShowManyBytes);
                                                          B->_Sdata.Adopt(bytesRead/4,arr);
                                                        }
                                               }
         B->_dataNotYetRead= 0;
       }
}

